from collections import Counter
from typing import Dict

import pandas as pd

from tangled_up_in_unicode import category, category_long, script, block, block_abbr


def get_character_counts(series) -> Dict[str, int]:
    """Function to return the character counts

    Args:
        series: the Series to process

    Returns:
        A dict with character counts
    """
    return dict(Counter(series.str.cat()))


def text_summary(series: pd.Series) -> dict:
    """

    Args:
        series: series to summarize

    Returns:

    """
    # Distribution of length
    summary = {"length": series.map(lambda x: len(str(x))).value_counts().to_dict()}

    # Unicode Character Summaries (category and script name)
    character_counts = get_character_counts(series)

    summary["category_short_values"] = {
        key: category(key) for key in character_counts.keys()
    }
    summary["category_alias_values"] = {
        key: category_long(value)
        for key, value in summary["category_short_values"].items()
    }
    summary["script_values"] = {key: script(key) for key in character_counts.keys()}
    summary["block_values"] = {key: block(key) for key in character_counts.keys()}
    summary["block_alias_values"] = {
        key: block_abbr(value) for key, value in summary["block_values"].items()
    }

    return summary



import pytest

from tests.series import get_series, get_contains_map


def all_series_included(series_list, series_map):
    """Check that all names are indeed used"""
    used_names = set([name for names in series_map.values() for name in names])
    names = set([series.name for series in series_list])
    if not names == used_names:
        raise ValueError(
            "Not all series are used {unused}".format(unused=names ^ used_names)
        )


def pytest_generate_tests(metafunc):
    _test_suite = get_series()
    if metafunc.function.__name__ == "test_contains":
        _series_map = get_contains_map()

        all_series_included(_test_suite, _series_map)

        argsvalues = []
        for item in _test_suite:
            for type, series_list in _series_map.items():
                args = {
                    "id": "{item_name} x {item_type}".format(
                        item_name=item.name, item_type=type
                    )
                }

                member = item.name in series_list
                argsvalues.append(pytest.param(item, type, member, **args))

        metafunc.parametrize(
            argnames=["series", "type", "member"], argvalues=argsvalues
        )


def test_contains(series, type, member):
    assert member == (series in type)



import pandas as pd

from visions.functional import detect_series_type
from visions.typesets import StandardSet
from visions.types import Categorical, Boolean

from examples.data_analysis.categorical import Category

ts = StandardSet()
ts -= Boolean
ts -= Categorical
ts += Category

s1 = pd.Series(["A", "B", "C"] * 1000, dtype="category")
print(s1 in Category)
print(detect_series_type(s1, ts))

s2 = pd.Series([True, False] * 1000)
print(s2 in Category)
print(detect_series_type(s2, ts))



import pandas as pd


def existing_path_summary(series: pd.Series) -> dict:
    """

    Args:
        series: series to summarize

    Returns:

    """
    summary = {"file_sizes": series.map(lambda x: x.stat().st_size)}
    return summary



from pathlib import Path

from visions.types import Integer, ExistingPath, String, DateTime
from visions.application.summaries.complete_summary import CompleteSummary

summaries_dir = Path("summaries/")
summaries_dir.mkdir(exist_ok=True)

summary = CompleteSummary()

# For a typeset
summary.plot(summaries_dir / "summary_complete.svg")

# For specific types
summary.plot(summaries_dir / "summary_integer.svg", Integer)
summary.plot(summaries_dir / "summary_string.svg", String)
summary.plot(summaries_dir / "summary_datetime.svg", DateTime)
summary.plot(summaries_dir / "summary_existing_path.svg", ExistingPath)



from visions.types import *
from visions.typesets import CompleteSet
from visions.application.summaries.summary import Summary
from visions.application.summaries import *


class CompleteSummary(Summary):
    def __init__(self):
        type_summary_ops = {
            Boolean: [],
            Categorical: [category_summary, unique_summary],
            Complex: [
                infinite_summary,
                numerical_basic_summary,
                unique_summary_complex,
            ],
            DateTime: [range_summary, unique_summary],
            Date: [],
            ExistingPath: [existing_path_summary, path_summary, text_summary],
            Float: [infinite_summary, numerical_summary, zero_summary, unique_summary],
            Geometry: [],
            ImagePath: [],
            Integer: [
                infinite_summary,
                numerical_summary,
                zero_summary,
                unique_summary,
            ],
            Object: [unique_summary],
            Path: [path_summary, text_summary],
            String: [text_summary, unique_summary],
            Time: [],
            TimeDelta: [],
            UUID: [],
            URL: [url_summary, unique_summary],
            Generic: [base_summary, missing_summary],
        }
        super().__init__(type_summary_ops, CompleteSet())



import imghdr
from pathlib import Path
from typing import Sequence

import pandas as pd

from visions.relations import IdentityRelation, TypeRelation
from visions.types import VisionsBaseType


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import ExistingPath

    relations = [IdentityRelation(cls, ExistingPath)]
    return relations


class ImagePath(VisionsBaseType):
    """**Image Path** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([Path('/home/user/file.png'), Path('/home/user/test2.jpg')])
        >>> x in visions.ImagePath
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        return all(
            isinstance(p, Path) and p.exists() and imghdr.what(p) for p in series
        )



from typing import Callable

import pandas as pd


def nullable_series_contains(fn: Callable) -> Callable:
    def inner(cls, series: pd.Series) -> bool:
        if series.hasnans:
            series = series.dropna()
            if series.empty:
                return False

        return fn(cls, series)

    return inner



import pandas as pd
import numpy as np

from visions.application.summaries.functional import (
    summarize_frame,
    summarize_series,
    summarize,
)
from visions.application.summaries import CompleteSummary
from visions.types import String, Integer


def test_summarize_frame():
    df = pd.DataFrame(
        {
            "Brand": ["Honda Civic", "Toyota Corolla", "Ford Focus", "Audi A4"],
            "Price": [22000, 25000, 27000, 35000],
        },
        columns=["Brand", "Price"],
    )
    summary = summarize_frame(
        df, {"Brand": String, "Price": Integer}, CompleteSummary()
    )
    summary.pop("type_counts")
    assert summary == {
        "n_observations": 4,
        "n_variables": 2,
        "memory_size": 430,
        "na_count": 0,
        "n_vars_missing": 0,
    }


def test_summarize_series():
    brand_series = pd.Series(["Honda Civic", "Toyota Corolla", "Ford Focus", "Audi A4"])

    summary = summarize_series(brand_series, String, CompleteSummary())
    assert summary == {
        "n_unique": 4,
        "length": {7: 1, 14: 1, 11: 1, 10: 1},
        "category_short_values": {
            "H": "Lu",
            "o": "Ll",
            "n": "Ll",
            "d": "Ll",
            "a": "Ll",
            " ": "Zs",
            "C": "Lu",
            "i": "Ll",
            "v": "Ll",
            "c": "Ll",
            "T": "Lu",
            "y": "Ll",
            "t": "Ll",
            "r": "Ll",
            "l": "Ll",
            "F": "Lu",
            "u": "Ll",
            "s": "Ll",
            "A": "Lu",
            "4": "Nd",
        },
        "category_alias_values": {
            "H": "Uppercase_Letter",
            "o": "Lowercase_Letter",
            "n": "Lowercase_Letter",
            "d": "Lowercase_Letter",
            "a": "Lowercase_Letter",
            " ": "Space_Separator",
            "C": "Uppercase_Letter",
            "i": "Lowercase_Letter",
            "v": "Lowercase_Letter",
            "c": "Lowercase_Letter",
            "T": "Uppercase_Letter",
            "y": "Lowercase_Letter",
            "t": "Lowercase_Letter",
            "r": "Lowercase_Letter",
            "l": "Lowercase_Letter",
            "F": "Uppercase_Letter",
            "u": "Lowercase_Letter",
            "s": "Lowercase_Letter",
            "A": "Uppercase_Letter",
            "4": "Decimal_Number",
        },
        "script_values": {
            "H": "Latin",
            "o": "Latin",
            "n": "Latin",
            "d": "Latin",
            "a": "Latin",
            " ": "Common",
            "C": "Latin",
            "i": "Latin",
            "v": "Latin",
            "c": "Latin",
            "T": "Latin",
            "y": "Latin",
            "t": "Latin",
            "r": "Latin",
            "l": "Latin",
            "F": "Latin",
            "u": "Latin",
            "s": "Latin",
            "A": "Latin",
            "4": "Common",
        },
        "block_values": {
            "H": "Basic Latin",
            "o": "Basic Latin",
            "n": "Basic Latin",
            "d": "Basic Latin",
            "a": "Basic Latin",
            " ": "Basic Latin",
            "C": "Basic Latin",
            "i": "Basic Latin",
            "v": "Basic Latin",
            "c": "Basic Latin",
            "T": "Basic Latin",
            "y": "Basic Latin",
            "t": "Basic Latin",
            "r": "Basic Latin",
            "l": "Basic Latin",
            "F": "Basic Latin",
            "u": "Basic Latin",
            "s": "Basic Latin",
            "A": "Basic Latin",
            "4": "Basic Latin",
        },
        "block_alias_values": {
            "H": "ASCII",
            "o": "ASCII",
            "n": "ASCII",
            "d": "ASCII",
            "a": "ASCII",
            " ": "ASCII",
            "C": "ASCII",
            "i": "ASCII",
            "v": "ASCII",
            "c": "ASCII",
            "T": "ASCII",
            "y": "ASCII",
            "t": "ASCII",
            "r": "ASCII",
            "l": "ASCII",
            "F": "ASCII",
            "u": "ASCII",
            "s": "ASCII",
            "A": "ASCII",
            "4": "ASCII",
        },
        "frequencies": {
            "Audi A4": 1,
            "Honda Civic": 1,
            "Ford Focus": 1,
            "Toyota Corolla": 1,
        },
        "n_records": 4,
        "memory_size": 398,
        "dtype": np.object_,
        "types": {"str": 4},
        "na_count": 0,
    }

    price_series = pd.Series([22000, 25000, 27000, 35000])

    summary = summarize_series(price_series, Integer, CompleteSummary())
    assert summary == {
        "inf_count": 0,
        "mean": 27250.0,
        "std": 5560.275772537426,
        "var": 30916666.666666668,
        "max": 35000.0,
        "min": 22000.0,
        "median": 26000.0,
        "kurt": 1.8192544372679649,
        "skew": 1.19978923754086,
        "sum": 109000.0,
        "mad": 3875.0,
        "quantile_5": 22450.0,
        "quantile_25": 24250.0,
        "quantile_50": 26000.0,
        "quantile_75": 29000.0,
        "quantile_95": 33800.0,
        "iqr": 4750.0,
        "range": 13000.0,
        "cv": 0.20404681734082297,
        "n_zeros": 0,
        "n_unique": 4,
        "frequencies": {27000: 1, 35000: 1, 25000: 1, 22000: 1},
        "n_records": 4,
        "memory_size": 160,
        "dtype": np.int64,
        "types": {"int": 4},
        "na_count": 0,
        "monotonic_decrease": False,
        "monotonic_increase": True,
    }


def test_summarize():
    df = pd.DataFrame(
        {
            "Brand": ["Honda Civic", "Toyota Corolla", "Ford Focus", "Audi A4"],
            "Price": [22000, 25000, 27000, 35000],
        },
        columns=["Brand", "Price"],
    )
    summary = summarize(df, {"Brand": String, "Price": Integer}, CompleteSummary())

    assert summary["types"] == {"Brand": String, "Price": Integer}



import pandas as pd


def unique_summary(series: pd.Series) -> dict:
    """

    Args:
        series: series to summarize

    Returns:

    """
    summary = {}
    summary.update({"n_unique": series.nunique()})
    return summary


def unique_summary_complex(series: pd.Series) -> dict:
    """

    Args:
        series: series to summarize

    Returns:

    """
    summary = {}
    # Until complex bug is fixed:
    # https://github.com/pandas-dev/pandas/issues/17927
    # https://github.com/pandas-dev/pandas/pull/27599
    n_unique = len(set(series[series.notna()].values))
    summary.update({"n_unique": n_unique})
    return summary



import warnings


def suppress_warnings(func):
    def inner(*args, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return func(*args, **kwargs)

    return inner



"""Core functionality"""
from visions import utils

from visions.dtypes.boolean import BoolDtype

from visions import types, typesets

from visions.types import *
from visions.typesets import *

from visions.functional import (
    cast_frame,
    infer_frame_type,
    detect_frame_type,
    cast_series,
    infer_series_type,
    detect_series_type,
)



import numpy as np
import pandas as pd


def infinite_summary(series: pd.Series) -> dict:
    """Counts the number of infinite values.

    Args:
        series: series to summarize

    Returns:
        A dict containing `inf_count`.
    """
    summary = {}
    mask = (~np.isfinite(series)) & series.notnull()
    summary["inf_count"] = mask.values.sum()
    return summary



from pathlib import Path
from typing import Dict, List, Any

import pandas as pd

from visions.utils.images.image_utils import (
    open_image,
    is_image_truncated,
    extract_exif,
    hash_image,
)


def count_duplicate_hashes(image_descriptions: dict) -> int:
    """

    Args:
        image_descriptions:

    Returns:

    """
    counts = pd.Series(
        [x["hash"] for x in image_descriptions if "hash" in x]
    ).value_counts()
    return counts.sum() - len(counts)


def get_exit_values() -> Dict[str, List]:
    return {}


def extract_exif_series(image_exifs: list) -> dict:
    """

    Args:
        image_exifs:

    Returns:

    """
    exif_keys = []
    exif_values = get_exit_values()

    for image_exif in image_exifs:
        # Extract key
        exif_keys.extend(list(image_exif.keys()))

        # Extract values per key
        for exif_key, exif_val in image_exif.items():
            if exif_key not in exif_values:
                exif_values[exif_key] = []

            exif_values[exif_key].append(exif_val)

    series = {"exif_keys": pd.Series(exif_keys).value_counts().to_dict()}

    for k, v in exif_values.items():
        series[k] = pd.Series(v).value_counts().to_dict()

    return series


def get_information() -> Dict[Any, Any]:
    return {}


def extract_image_information(path: Path) -> dict:
    """Extracts all image information per file, as opening files is slow

    Args:
        path: Path to the image

    Returns:
        A dict containing image information
    """
    information = get_information()
    image = open_image(path)
    information["opened"] = image is not None
    if image is not None:
        information["truncated"] = is_image_truncated(image)
        if not information["truncated"]:
            information["size"] = image.size
            information["exif"] = extract_exif(image)
            information["hash"] = hash_image(image)
        # else:
        #     print(image.size)
    return information


def image_summary(series: pd.Series) -> dict:
    """

    Args:
        series: series to summarize

    Returns:

    """
    from visions.utils.images.image_utils import (
        open_image,
        is_image_truncated,
        extract_exif,
        hash_image,
    )

    image_information = series.apply(extract_image_information)
    summary = {
        "n_duplicate_hash": count_duplicate_hashes(image_information),
        "n_truncated": sum(
            [1 for x in image_information if "truncated" in x and x["truncated"]]
        ),
    }

    exif_series = extract_exif_series(
        [x["exif"] for x in image_information if "exif" in x]
    )
    summary["exif_keys_counts"] = exif_series["exif_keys"]

    image_shapes = pd.Series(
        [x["size"] for x in image_information if "size" in x], name="image_shape"
    )
    summary["image_shape_counts"] = image_shapes.value_counts().to_dict()

    return summary



from typing import Sequence

import numpy as np
import pandas as pd
import pandas.api.types as pdt

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.relations.string_to_bool import get_boolean_coercions
from visions.types import VisionsBaseType
from visions.utils.coercion.test_utils import coercion_map_test, coercion_map


def to_bool(series: pd.Series) -> pd.Series:
    if series.isin({True, False}).all():
        return series.astype(bool)
    elif series.isin({True, False, None, np.nan}).all():
        return series.astype("Bool")
    else:
        unsupported_values = series[~series.isin({True, False, None, np.nan})].unique()
        raise ValueError(
            "Values not supported {unsupported_values}".format(
                unsupported_values=unsupported_values
            )
        )


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Generic, String, Integer, Object

    relations = [
        IdentityRelation(cls, Generic),
        InferenceRelation(
            cls,
            String,
            relationship=lambda s: coercion_map_test(cls.string_coercions)(
                s.str.lower()
            ),
            transformer=lambda s: to_bool(
                coercion_map(cls.string_coercions)(s.str.lower())
            ),
        ),
        InferenceRelation(
            cls,
            Integer,
            relationship=lambda s: s.isin({0, 1, np.nan}).all(),
            transformer=to_bool,
        ),
        InferenceRelation(
            cls,
            Object,
            relationship=lambda s: s.apply(type).isin([type(None), bool]).all(),
            transformer=to_bool,
        ),
    ]
    return relations


class Boolean(VisionsBaseType):
    """**Boolean** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([True, False, False, True])
        >>> x in visions.Boolean
        True

        >>> x = pd.Series([True, False, None])
        >>> x in visions.Boolean
        True
    """

    string_coercions = get_boolean_coercions("en")

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        return not pdt.is_categorical_dtype(series) and pdt.is_bool_dtype(series)

    @classmethod
    def make_string_coercion(cls, type_name, string_coercions):
        @classmethod
        def get_relations(cls):
            return _get_relations(cls)

        return type(
            "{name}[{type_name}]".format(name=cls.__name__, type_name=type_name),
            (cls,),
            {
                "string_coercions": string_coercions,
                "get_relations": get_relations,
                "contains_op": cls.contains_op,
                "make_string_coercion": cls.make_string_coercion,
            },
        )



import os

import pandas as pd


def path_summary(series: pd.Series) -> dict:
    """

    Args:
        series: series to summarize

    Returns:

    """
    summary = {
        "common_prefix": (os.path.commonprefix(list(series)) or "No common prefix"),
        "stem_counts": series.map(lambda x: x.stem).value_counts().to_dict(),
        "suffix_counts": (series.map(lambda x: x.suffix).value_counts().to_dict()),
        "name_counts": series.map(lambda x: x.name).value_counts().to_dict(),
        "parent_counts": (series.map(lambda x: x.parent).value_counts().to_dict()),
        "anchor_counts": (series.map(lambda x: x.anchor).value_counts().to_dict()),
    }

    return summary



from pathlib import Path

from visions.typesets import CompleteSet, StandardSet, GeometrySet

# Windows Note
# Tip for Python3/64-bit compatible version of pygraphviz
# https://github.com/CristiFati/Prebuilt-Binaries/raw/master/Windows/PyGraphviz/pygraphviz-1.5-cp37-cp37m-win_amd64.whl


typesets_dir = Path("typesets/")
typesets_dir.mkdir(exist_ok=True)

# Initialize typeset
for name, tsc in [
    ("typeset_complete", CompleteSet()),
    ("typeset_geometry", GeometrySet()),
    ("typeset_standard", StandardSet()),
]:
    # Write graph to dot
    tsc.output_graph(typesets_dir / "{name}.dot".format(name=name))

    # Plot the graph (svg)
    tsc.output_graph(typesets_dir / "{name}.svg".format(name=name))
    tsc.output_graph(typesets_dir / "{name}_base.svg".format(name=name), base_only=True)

    # Plot the graph (png)
    tsc.output_graph(typesets_dir / "{name}.png".format(name=name), dpi=150)



from typing import Sequence, List

import pandas.api.types as pdt
import pandas as pd
import numpy as np

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types import VisionsBaseType
from visions.utils.coercion import test_utils


def to_int(series: pd.Series) -> pd.Series:
    try:
        return series.astype(int)
    except ValueError:
        return series.astype("Int64")


def float_is_int(series: pd.Series) -> bool:
    def check_equality(series):
        if series.empty or not np.isfinite(series).all():
            return False
        return series.eq(series.astype(int)).all()

    return check_equality(series.dropna() if series.hasnans else series)


def test_string_is_int(series) -> bool:
    coerced_series = test_utils.option_coercion_evaluator(string_to_int)(series)
    return coerced_series is not None and coerced_series in Integer


def string_to_int(series: pd.Series) -> pd.Series:
    # if any("," in x for x in series.dropna()):
    #     series = series.str.replace(",", "")

    return to_int(series)


def _get_relations(cls) -> List[TypeRelation]:
    from visions.types import Generic, Float

    relations = [
        IdentityRelation(cls, Generic),
        InferenceRelation(cls, Float, relationship=float_is_int, transformer=to_int),
    ]
    return relations


class Integer(VisionsBaseType):
    """**Integer** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([1, 2, 3])
        >>> x in visions.Integer
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        return pdt.is_signed_integer_dtype(series)



from typing import Sequence

import pandas as pd
import pandas.api.types as pdt

from visions.relations import IdentityRelation, TypeRelation
from visions.types.type import VisionsBaseType


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Object

    relations = [IdentityRelation(cls, Object)]
    return relations


class String(VisionsBaseType):
    """**String** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series(['rubin', 'carter', 'champion'])
        >>> x in visions.String
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        # TODO: without the object check this passes string categories... is there a better way?
        if not pdt.is_object_dtype(series):
            return False
        elif series.hasnans:
            series = series.dropna()
            if series.empty:
                return False

        return all(type(v) is str for v in series)



from typing import Sequence

import pandas.api.types as pdt
import pandas as pd

from visions.relations import IdentityRelation, TypeRelation
from visions.types.type import VisionsBaseType


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import DateTime

    relations = [IdentityRelation(cls, DateTime)]
    return relations


class Time(VisionsBaseType):
    """**Time** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([pd.datetime(2017, 3, 5), pd.datetime(2019, 12, 4)])
        >>> x in visions.Time
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        if not pdt.is_datetime64_any_dtype(series):
            return False

        temp_series = series.dropna().dt

        time_val_map = {"day": 1, "month": 1, "year": 1}
        return all(
            getattr(temp_series, time_part).eq(val).all()
            for time_part, val in time_val_map.items()
        )



""" Utilities suite for visions """
from visions.utils.coercion import test_utils

# from visions.utils.images import image_utils
from visions.utils.monkeypatches import pathlib_patch, imghdr_patch
from visions.utils.warning_handling import suppress_warnings



import pandas.api.types as pdt
import pandas as pd
import numpy as np
from typing import Sequence, List

from visions.types import String, Integer, Object, VisionsBaseType
from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.utils.coercion.test_utils import coercion_map_test, coercion_map


def to_category(series: pd.Series) -> pd.Series:
    if series.isin({True, False}).all():
        return series.astype(bool)
    elif series.isin({True, False, None, np.nan}).all():
        return series.astype("Bool")
    else:
        unsupported_values = series[~series.isin({True, False, None, np.nan})].unique()
        raise ValueError(
            "Values not supported {unsupported_values}".format(
                unsupported_values=unsupported_values
            )
        )


def _get_relations(cls) -> List[TypeRelation]:
    from visions.types import Generic

    relations = [
        IdentityRelation(cls, Generic),
        InferenceRelation(
            cls,
            String,
            relationship=lambda s: coercion_map_test(cls.string_coercions)(
                s.str.lower()
            ),
            transformer=lambda s: to_category(
                coercion_map(cls.string_coercions)(s.str.lower())
            ),
        ),
        InferenceRelation(
            cls,
            Integer,
            relationship=lambda s: s.isin({0, 1, np.nan}).all(),
            transformer=to_category,
        ),
        InferenceRelation(
            cls,
            Object,
            relationship=lambda s: s.apply(type).isin([type(None), bool]).all(),
            transformer=to_category,
        ),
    ]
    return relations


class Category(VisionsBaseType):
    """**Categorical** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([True, False, 1], dtype='category')
        >>> x in visions.Category
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        return pdt.is_categorical_dtype(series) or pdt.is_bool_dtype(series)



import sys

import pytest

from visions.typesets import CompleteSet
from tests.series import get_series, get_convert_map


def all_relations_tested(series_map):
    typeset = CompleteSet()

    # Convert data structure for mapping
    series_map_lookup = {}
    for map_to_type, map_from_type, items in series_map:
        try:
            series_map_lookup[map_to_type][map_from_type] = items
        except KeyError:
            series_map_lookup[map_to_type] = {map_from_type: items}

    missing_relations = set()
    for node in typeset.types:
        for relation in node.relations:
            from_type, to_type = relation.related_type, relation.type
            if relation.inferential and (
                to_type not in series_map_lookup
                or from_type not in series_map_lookup[to_type]
                or len(series_map_lookup[to_type][from_type]) == 0
            ):
                missing_relations.add(str(relation))

    if len(missing_relations) > 0:
        raise ValueError(
            "Not all inferential relations are tested {missing_relations}".format(
                missing_relations=missing_relations
            )
        )


def pytest_generate_tests(metafunc):
    _test_suite = get_series()
    if metafunc.function.__name__ == "test_relations":
        _series_map = get_convert_map()

        all_relations_tested(_series_map)

        argsvalues = []
        for item in _test_suite:
            for source_type, relation_type, series_list in _series_map:
                if item in relation_type:
                    args = {
                        "id": "{name}: {relation_type} -> {source_type}".format(
                            name=item.name,
                            relation_type=relation_type,
                            source_type=source_type,
                        )
                    }
                    member = item.name in series_list
                    argsvalues.append(
                        pytest.param(source_type, relation_type, item, member, **args)
                    )

        metafunc.parametrize(
            argnames=["source_type", "relation_type", "series", "member"],
            argvalues=argsvalues,
        )
    if metafunc.function.__name__ in [
        "test_consistency",
        "test_side_effects",
        "test_multiple_inference",
    ]:
        argsvalues = []
        for series in _test_suite:
            args = {"id": str(series.name)}
            argsvalues.append(pytest.param(series, **args))

        metafunc.parametrize(argnames=["series"], argvalues=argsvalues)


def test_relations(source_type, relation_type, series, member):
    relation_gen = (
        rel for rel in source_type.relations if rel.related_type == relation_type
    )
    relation = next(relation_gen)

    is_relation = relation.is_relation(series)

    if not member:
        assert (
            not is_relation
        ), f"{source_type}, {relation}, {member}, {series.name}, {series[0]}"
    else:
        assert is_relation
    if relation.is_relation(series):
        cast_series = relation.transform(series)

        assert (
            cast_series in source_type
        ), "Relationship {relation} cast {series_values} to {cast_values}".format(
            relation=relation,
            series_values=series.values,
            cast_values=cast_series.values,
        )


def test_consistency(series):
    typeset = CompleteSet()

    if (
        series.name
        in ["timedelta_series_nat", "date_series_nat", "timestamp_series_nat"]
        and sys.version_info.major == 3
        and sys.version_info.minor == 7
    ):
        pytest.skip("unsupported configuration")

    initial_type = typeset.detect_series_type(series.copy(deep=True))
    converted_type = typeset.infer_series_type(series.copy(deep=True))

    if initial_type != converted_type:
        converted_series = typeset.cast_series(series.copy(deep=True))
        assert series.dtype.kind != converted_series.dtype.kind or not (
            (
                converted_series.eq(series) ^ (converted_series.isna() & series.isna())
            ).all()
        )
    else:
        converted_series = typeset.cast_series(series.copy(deep=True))
        assert (
            converted_series.eq(series) ^ (converted_series.isna() & series.isna())
        ).all()


def test_side_effects(series):
    reference = series.copy()

    typeset = CompleteSet()
    typeset.detect_series_type(series)
    typeset.infer_series_type(series)

    # Check if NaN mask is equal
    assert series.notna().eq(reference.notna()).all()
    # Check if NonNaN values are equal
    assert series[series.notna()].eq(reference[reference.notna()]).all()


def test_multiple_inference(series):
    """
    Notes:
        Copy to prevent possible side effects only for testing.
    """
    ts = CompleteSet()

    inferred_type = ts.infer_series_type(series)

    series_convert = ts.cast_series(series.copy(deep=True))

    initial_type_after_convert = ts.detect_series_type(series_convert.copy(deep=True))
    assert inferred_type == initial_type_after_convert

    series_convert2 = ts.cast_series(series_convert.copy(deep=True))

    inferred_type_after_convert = ts.detect_series_type(series_convert2.copy(deep=True))
    assert initial_type_after_convert == inferred_type_after_convert

    assert series_convert.isna().eq(series_convert2.isna()).all()
    assert (
        series_convert[series_convert.notna()]
        .eq(series_convert2[series_convert2.notna()])
        .all()
    )



from typing import Sequence

import pandas.api.types as pdt
import pandas as pd

from visions.relations import IdentityRelation, TypeRelation
from visions.types import VisionsBaseType


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Generic

    relations = [IdentityRelation(cls, Generic)]
    return relations


class Categorical(VisionsBaseType):
    """**Categorical** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([True, False, 1], dtype='category')
        >>> x in visions.Categorical
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        return pdt.is_categorical_dtype(series)



import pandas as pd


def url_summary(series: pd.Series) -> dict:
    """

    Args:
        series: series to summarize

    Returns:

    """
    summary = {}

    keys = ["scheme", "netloc", "path", "query", "fragment"]
    url_parts = dict(zip(keys, zip(*series)))
    for name, part in url_parts.items():
        summary["{}_counts".format(name.lower())] = (
            pd.Series(part).value_counts().to_dict()
        )

    return summary



import numpy as np
import pandas as pd


def numerical_summary(series: pd.Series) -> dict:
    """

    Args:
        series: series to summarize

    Returns:

    """
    aggregates = [
        "mean",
        "std",
        "var",
        "max",
        "min",
        "median",
        "kurt",
        "skew",
        "sum",
        "mad",
    ]
    summary = series.agg(aggregates).to_dict()

    quantiles = [0.05, 0.25, 0.5, 0.75, 0.95]
    for percentile, value in series.quantile(quantiles).to_dict().items():
        summary["quantile_{:d}".format(int(percentile * 100))] = value
    summary["iqr"] = summary["quantile_75"] - summary["quantile_25"]

    summary["range"] = summary["max"] - summary["min"]
    summary["cv"] = summary["std"] / summary["mean"] if summary["mean"] else np.NaN

    summary["monotonic_increase"] = series.is_monotonic_increasing
    summary["monotonic_decrease"] = series.is_monotonic_decreasing

    # TODO: need access to n_unique
    # summary['monotonic_increase_strict'] = summary['monotonic_increase'] and summary['unique']
    # summary['monotonic_decrease_strict'] = summary['monotonic_increase'] and summary['unique']

    # TODO: only calculations for histogram, not the plotting
    # summary['image'] = plotting.histogram(series)
    return summary



import datetime
import uuid
from ipaddress import IPv4Address
from pathlib import PureWindowsPath, PurePosixPath
from urllib.parse import urlparse
import pandas as pd
import numpy as np
from shapely import wkt

from visions.types import (
    Boolean,
    Categorical,
    Complex,
    Count,
    Date,
    DateTime,
    ExistingPath,
    Float,
    Generic,
    Geometry,
    ImagePath,
    Integer,
    IPAddress,
    Object,
    Ordinal,
    Path,
    String,
    Time,
    TimeDelta,
    URL,
    UUID,
)


def get_series():
    return [
        # Int Series
        pd.Series([1, 2, 3], name="int_series"),
        pd.Series(range(10), name="int_range"),
        pd.Series([1, 2, 3], name="Int64_int_series", dtype="Int64"),
        pd.Series([1, 2, 3, np.nan], name="Int64_int_nan_series", dtype="Int64"),
        pd.Series([1, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0], name="int_series_boolean"),
        # Count
        pd.Series(np.array([1, 2, 3, 4], dtype=np.uint32), name="np_uint32"),
        # Categorical
        pd.Series([1, 2, 3], name="categorical_int_series", dtype="category"),
        pd.Series(
            pd.Categorical(
                ["A", "B", "C", "C", "B", "A"],
                categories=["A", "B", "C"],
                ordered=False,
            ),
            name="categorical_char",
        ),
        pd.Series([1.0, 2.0, 3.1], dtype="category", name="categorical_float_series"),
        pd.Series(
            ["Georgia", "Sam"], dtype="category", name="categorical_string_series"
        ),
        pd.Series(
            [np.complex(0, 0), np.complex(1, 2), np.complex(3, -1)],
            name="categorical_complex_series",
            dtype="category",
        ),
        # Ordinal
        pd.Series(
            pd.Categorical(
                ["A", "B", "C", "C", "B", "A"], categories=["A", "B", "C"], ordered=True
            ),
            name="ordinal",
        ),
        # Float Series
        pd.Series([1.0, 2.1, 3.0], name="float_series"),
        pd.Series([1.0, 2.5, np.nan], name="float_nan_series"),
        pd.Series([1.0, 2.0, 3.0, 4.0], name="float_series2"),
        pd.Series(np.array([1.2, 2, 3, 4], dtype=np.float), name="float_series3"),
        pd.Series([1, 2, 3.05, 4], dtype=np.float, name="float_series4"),
        pd.Series([np.nan, 1.2], name="float_series5"),
        pd.Series([np.nan, 1.1], dtype=np.single, name="float_series6"),
        pd.Series([np.inf, np.NINF, np.PINF, 1000000.0, 5.5], name="float_with_inf"),
        pd.Series([np.inf, np.NINF, np.Infinity, np.PINF], name="inf_series"),
        pd.Series([1, 2, np.nan], name="int_nan_series"),
        # Nan Series
        pd.Series([np.nan], name="nan_series"),
        pd.Series([np.nan, np.nan, np.nan, np.nan], name="nan_series_2"),
        # String Series
        pd.Series(["Patty", "Valentine"], name="string_series"),
        pd.Series(["1941-05-24", "13/10/2016"], name="timestamp_string_series"),
        pd.Series(["mack", "the", "finger"], name="string_unicode_series"),
        pd.Series(
            np.array(["upper", "hall"], dtype=np.unicode_),
            name="string_np_unicode_series",
        ),
        pd.Series(["1.0", "2.0", np.nan], name="string_num_nan"),
        pd.Series(["1,000.0", "2.1", np.nan], name="string_with_sep_num_nan"),
        pd.Series(["1.0", "2.0", "3.0"], name="string_num"),
        pd.Series(["1.0", "45.67", np.nan], name="string_flt_nan"),
        pd.Series(["1.0", "45.67", "3.5"], name="string_flt"),
        pd.Series(
            ["POINT (-92 42)", "POINT (-92 42.1)", "POINT (-92 42.2)"],
            name="geometry_string_series",
        ),
        pd.Series(
            [
                "I was only robbing the register,",
                "I hope you understand",
                "One of us had better call up the cops",
                "In the hot New Jersey night",
                np.nan,
            ],
            name="string_str_nan",
        ),
        pd.Series(["True", "False", None], name="string_bool_nan"),
        pd.Series(range(20), name="int_str_range").astype("str"),
        pd.Series(["1937-05-06", "20/4/2014"], name="string_date"),
        pd.Series(
            [
                "http://www.cwi.nl:80/%7Eguido/Python.html",
                "https://github.com/pandas-profiling/pandas-profiling",
            ],
            name="str_url",
        ),
        pd.Series(
            [r"C:\\home\\user\\file.txt", r"C:\\home\\user\\test2.txt"],
            name="path_series_windows_str",
        ),
        pd.Series(
            [r"/home/user/file.txt", r"/home/user/test2.txt"],
            name="path_series_linux_str",
        ),
        # Bool Series
        pd.Series([True, False], name="bool_series"),
        pd.Series([True, False, None], name="bool_nan_series"),
        pd.Series([True, False, None], name="nullable_bool_series", dtype="Bool"),
        pd.Series([True, False, False, True], dtype=bool, name="bool_series2"),
        pd.Series(np.array([1, 0, 0, 1], dtype=np.bool), name="bool_series3"),
        # Complex Series
        pd.Series(
            [np.complex(0, 0), np.complex(1, 2), np.complex(3, -1)],
            name="complex_series",
        ),
        pd.Series(
            [
                np.complex(0, 0),
                np.complex(1, 2),
                np.complex(3, -1),
                np.complex(np.nan, np.nan),
            ],
            name="complex_series_nan",
        ),
        pd.Series(["(1+1j)", "(2+2j)", "(10+100j)"], name="str_complex"),
        pd.Series(
            [np.complex(0, 0), np.complex(1, 2), np.complex(3, -1), np.nan],
            name="complex_series_nan_2",
        ),
        pd.Series(
            [complex(0, 0), complex(1, 2), complex(3, -1), np.nan],
            name="complex_series_py_nan",
        ),
        pd.Series(
            [complex(0, 0), complex(1, 2), complex(3, -1)], name="complex_series_py"
        ),
        pd.Series(
            [np.complex(0, 0), np.complex(1, 0), np.complex(3, 0), np.complex(-1, 0)],
            name="complex_series_float",
        ),
        # Datetime Series
        pd.Series(
            [datetime.datetime(2017, 3, 5, 12, 2), datetime.datetime(2019, 12, 4)],
            name="timestamp_series",
        ),
        pd.Series(
            [
                datetime.datetime(2017, 3, 5),
                datetime.datetime(2019, 12, 4, 3, 2, 0),
                pd.NaT,
            ],
            name="timestamp_series_nat",
        ),
        pd.Series(
            [datetime.datetime(2017, 3, 5), datetime.datetime(2019, 12, 4), pd.NaT],
            name="date_series_nat",
        ),
        pd.Series(
            pd.date_range(
                start="2013-05-18 12:00:00",
                periods=2,
                freq="H",
                tz="Europe/Brussels",
                name="timestamp_aware_series",
            )
        ),
        pd.to_datetime(
            pd.Series(
                [
                    datetime.date(2011, 1, 1),
                    datetime.date(2012, 1, 2),
                    datetime.date(2013, 1, 1),
                ],
                name="datetime",
            )
        ),
        # Timedelta Series
        pd.Series([pd.Timedelta(days=i) for i in range(3)], name="timedelta_series"),
        pd.Series(
            [pd.Timedelta(days=i) for i in range(3)] + [pd.NaT],
            name="timedelta_series_nat",
        ),
        # Geometry Series
        pd.Series(
            [
                wkt.loads("POINT (-92 42)"),
                wkt.loads("POINT (-92 42.1)"),
                wkt.loads("POINT (-92 42.2)"),
            ],
            name="geometry_series",
        ),
        # Path Series
        pd.Series(
            [
                PurePosixPath("/home/user/file.txt"),
                PurePosixPath("/home/user/test2.txt"),
            ],
            name="path_series_linux",
        ),
        pd.Series(
            [
                PureWindowsPath("C:\\home\\user\\file.txt"),
                PureWindowsPath("C:\\home\\user\\test2.txt"),
            ],
            name="path_series_windows",
        ),
        # Url Series
        pd.Series(
            [
                urlparse("http://www.cwi.nl:80/%7Eguido/Python.html"),
                urlparse("https://github.com/dylan-profiling/hurricane"),
            ],
            name="url_series",
        ),
        pd.Series(
            [
                urlparse("http://www.cwi.nl:80/%7Eguido/Python.html"),
                urlparse("https://github.com/dylan-profiling/hurricane"),
                np.nan,
            ],
            name="url_nan_series",
        ),
        # UUID Series
        pd.Series(
            [
                uuid.UUID("0b8a22ca-80ad-4df5-85ac-fa49c44b7ede"),
                uuid.UUID("aaa381d6-8442-4f63-88c8-7c900e9a23c6"),
                uuid.UUID("00000000-0000-0000-0000-000000000000"),
            ],
            name="uuid_series",
        ),
        pd.Series(
            [
                "0b8a22ca-80ad-4df5-85ac-fa49c44b7ede",
                "aaa381d6-8442-4f63-88c8-7c900e9a23c6",
                "00000000-0000-0000-0000-000000000000",
            ],
            name="uuid_series_str",
        ),
        # Object Series
        pd.Series([[1, ""], [2, "Rubin"], [3, "Carter"]], name="mixed_list[str,int]"),
        pd.Series(
            [{"why": "did you"}, {"bring him": "in for he"}, {"aint": "the guy"}],
            name="mixed_dict",
        ),
        pd.Series(
            [pd.to_datetime, pd.to_timedelta, pd.read_json, pd.to_pickle],
            name="callable",
        ),
        pd.Series([pd, wkt, np], name="module"),
        pd.Series(["1.1", "2"], name="textual_float"),
        pd.Series(["1.1", "2", "NAN"], name="textual_float_nan"),
        # Empty
        pd.Series([], name="empty"),
        pd.Series([], name="empty_float", dtype=float),
        pd.Series([], name="empty_int64", dtype="Int64"),
        pd.Series([], name="empty_object", dtype="object"),
        pd.Series([], name="empty_bool", dtype=bool),
        # IP
        pd.Series([IPv4Address("127.0.0.1"), IPv4Address("127.0.0.1")], name="ip"),
        pd.Series(["127.0.0.1", "127.0.0.1"], name="ip_str"),
    ]


def get_contains_map():
    series_map = {
        Integer: [
            "int_series",
            "Int64_int_series",
            "int_range",
            "Int64_int_nan_series",
            "int_series_boolean",
        ],
        Count: ["np_uint32"],
        Path: ["path_series_linux", "path_series_windows"],
        URL: ["url_series", "url_nan_series"],
        Float: [
            "float_series",
            "float_series2",
            "float_series3",
            "float_series4",
            "inf_series",
            "nan_series",
            "float_nan_series",
            "float_series5",
            "int_nan_series",
            "nan_series_2",
            "float_with_inf",
            "float_series6",
        ],
        Categorical: [
            "categorical_int_series",
            "categorical_float_series",
            "categorical_string_series",
            "categorical_complex_series",
            "categorical_char",
            "ordinal",
        ],
        Boolean: [
            "bool_series",
            "bool_series2",
            "bool_series3",
            "nullable_bool_series",
        ],
        Complex: [
            "complex_series",
            "complex_series_py",
            "complex_series_nan",
            "complex_series_py_nan",
            "complex_series_nan_2",
            "complex_series_float",
        ],
        DateTime: [
            "timestamp_series",
            "timestamp_aware_series",
            "datetime",
            "timestamp_series_nat",
            "date_series_nat",
        ],
        Date: ["datetime", "date_series_nat"],
        TimeDelta: ["timedelta_series", "timedelta_series_nat"],
        String: [
            "timestamp_string_series",
            "string_with_sep_num_nan",
            "string_series",
            "geometry_string_series",
            "string_unicode_series",
            "string_np_unicode_series",
            "path_series_linux_str",
            "path_series_windows_str",
            "int_str_range",
            "string_date",
            "textual_float",
            "textual_float_nan",
            "ip_str",
            "string_flt",
            "string_num",
            "str_url",
            "string_str_nan",
            "string_num_nan",
            "string_bool_nan",
            "string_flt_nan",
            "str_complex",
            "uuid_series_str",
        ],
        Geometry: ["geometry_series"],
        IPAddress: ["ip"],
        Ordinal: ["ordinal"],
        UUID: ["uuid_series"],
    }

    series_map[Object] = (
        ["mixed_list[str,int]", "mixed_dict", "callable", "module", "bool_nan_series"]
        + series_map[String]
        + series_map[Geometry]
        + series_map[Path]
        + series_map[URL]
        + series_map[IPAddress]
        + series_map[UUID]
    )

    # Empty series
    all = ["empty", "empty_bool", "empty_float", "empty_int64", "empty_object"]
    for key, values in series_map.items():
        all += values
    series_map[Generic] = list(set(all))

    return series_map


def infer_series_type_map():
    return {
        "int_series": Integer,
        "categorical_int_series": Categorical,
        "int_nan_series": Integer,
        "Int64_int_series": Integer,
        "Int64_int_nan_series": Integer,
        "np_uint32": Count,
        "int_range": Integer,
        "float_series": Float,
        "float_nan_series": Float,
        "int_series_boolean": Boolean,
        "float_series2": Integer,
        "float_series3": Float,
        "float_series4": Float,
        "float_series5": Float,
        "float_series6": Float,
        "complex_series_float": Integer,
        "categorical_float_series": Categorical,
        "float_with_inf": Float,
        "inf_series": Float,
        "nan_series": Float,
        "nan_series_2": Float,
        "string_series": String,
        "categorical_string_series": Categorical,
        "timestamp_string_series": Date,
        "string_with_sep_num_nan": String,  # TODO: Introduce thousands separator
        "string_unicode_series": String,
        "string_np_unicode_series": String,
        "string_num_nan": Integer,
        "string_num": Integer,
        "string_flt_nan": Float,
        "string_flt": Float,
        "string_str_nan": String,
        "string_bool_nan": Boolean,
        "int_str_range": Integer,
        "string_date": Date,
        "str_url": URL,
        "bool_series": Boolean,
        "bool_nan_series": Boolean,
        "nullable_bool_series": Boolean,
        "bool_series2": Boolean,
        "bool_series3": Boolean,
        "complex_series": Complex,
        "complex_series_nan": Complex,
        "complex_series_nan_2": Complex,
        "complex_series_py_nan": Complex,
        "complex_series_py": Complex,
        "categorical_complex_series": Categorical,
        "timestamp_series": DateTime,
        "timestamp_series_nat": DateTime,
        "timestamp_aware_series": DateTime,
        "datetime": Date,
        "timedelta_series": TimeDelta,
        "timedelta_series_nat": TimeDelta,
        "geometry_string_series": Geometry,
        "geometry_series": Geometry,
        "path_series_linux": Path,
        "path_series_linux_str": Path,
        "path_series_windows": Path,
        "path_series_windows_str": Path,
        "url_series": URL,
        "url_nan_series": URL,
        "mixed_list[str,int]": Object,
        "mixed_dict": Object,
        "callable": Object,
        "module": Object,
        "textual_float": Float,
        "textual_float_nan": Float,
        "empty": Generic,
        "empty_object": Generic,
        "empty_float": Generic,
        "empty_bool": Generic,
        "empty_int64": Generic,
        "ip": IPAddress,
        "ip_str": IPAddress,
        "date_series_nat": Date,
        "categorical_char": Categorical,
        "ordinal": Ordinal,
        "str_complex": Complex,
        "uuid_series": UUID,
        "uuid_series_str": UUID,
    }


def get_convert_map():
    # Conversions in one single step
    series_map = [
        # Model type, Relation type
        (Integer, Float, ["int_nan_series", "float_series2"]),
        (Complex, String, ["str_complex"]),
        (
            Float,
            String,
            [
                "string_flt",
                "string_num_nan",
                "string_num",
                "string_flt_nan",
                "textual_float",
                "textual_float_nan",
                "int_str_range",
                # "string_with_sep_num_nan",
            ],
        ),
        (DateTime, String, ["timestamp_string_series", "string_date"]),
        (Geometry, String, ["geometry_string_series"]),
        (Boolean, String, ["string_bool_nan"]),
        (IPAddress, String, ["ip_str"]),
        (URL, String, ["str_url"]),
        (Path, String, ["path_series_windows_str", "path_series_linux_str"]),
        (Float, Complex, ["complex_series_float"]),
        (Boolean, Integer, ["int_series_boolean"]),
        (Boolean, Object, ["bool_nan_series"]),
        (UUID, String, ["uuid_series_str"]),
    ]

    return series_map



from typing import Sequence

import pandas.api.types as pdt
import pandas as pd

from visions.relations import IdentityRelation, TypeRelation
from visions.types import VisionsBaseType


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Generic

    relations = [IdentityRelation(cls, Generic)]
    return relations


class Object(VisionsBaseType):
    """**Object** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series(['a', 1, np.nan])
        >>> x in visions.Object
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        return pdt.is_object_dtype(series)



from typing import Sequence

import pandas.api.types as pdt
import numpy as np
import pandas as pd

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types import VisionsBaseType
from visions.utils.coercion import test_utils
from visions.utils.warning_handling import suppress_warnings


def test_string_is_float(series) -> bool:
    coerced_series = test_utils.option_coercion_evaluator(string_to_float)(series)
    return coerced_series is not None and coerced_series in Float


def string_to_float(series: pd.Series) -> pd.Series:
    # Slightly faster to check for the character if it's not present than to
    # attempt the replacement
    # if any("," in x for x in series.dropna()):
    #     series = series.str.replace(",", "")

    return to_float(series)


def test_is_float(series: pd.Series) -> bool:
    coerced_series = test_utils.option_coercion_evaluator(to_float)(series)
    return coerced_series is not None and coerced_series in Float


def to_float(series: pd.Series) -> pd.Series:
    return series.astype(float)


def _get_relations(cls) -> Sequence[TypeRelation]:
    from visions.types import Generic, String, Complex

    relations = [
        IdentityRelation(cls, Generic),
        InferenceRelation(
            cls, String, relationship=test_string_is_float, transformer=string_to_float
        ),
        InferenceRelation(
            cls,
            Complex,
            relationship=lambda s: all(np.imag(s.values) == 0),
            transformer=suppress_warnings(to_float),
        ),
    ]
    return relations


class Float(VisionsBaseType):
    """**Float** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = pd.Series([1.0, 2.5, 5.0, np.nan])
        >>> x in visions.Float
        True
    """

    @classmethod
    def get_relations(cls) -> Sequence[TypeRelation]:
        return _get_relations(cls)

    @classmethod
    def contains_op(cls, series: pd.Series) -> bool:
        return pdt.is_float_dtype(series)
